UNIT MyPrinting;

INTERFACE

USES
	Types,   QuickDraw, Events, Windows, Dialogs, Fonts, DiskInit, 
	Devices, TextEdit,  Traps,  Memory,  Processes, Scrap, ToolUtils, 
	OSUtils,   Menus, Palettes, GestaltEqu , Lists, TextUtils,
	QuickDrawText, Resources, Controls, Printing,
	
	Globals;

PROCEDURE SetFont( theFont : INTEGER; theSize : INTEGER; theStyle : Style );

FUNCTION GetSelectedFonts : IntegerListRef;
FUNCTION NoMoreRoom( srcRect : Rect;  height : INTEGER ) : BOOLEAN;
FUNCTION FontHeight : INTEGER;
FUNCTION NumberOfRows( thePage : Rect ) : INTEGER;

PROCEDURE DrawFont( thePort : TPPrPort; VAR thePage : Rect; theFont : INTEGER );
PROCEDURE ImagePage( thePort : TPPrPort; thePage : Rect; VAR fonts : IntegerListRef );
PROCEDURE PrintEngine( documentNum, pageNum	: INTEGER; fonts : IntegerListRef );
PROCEDURE GetPrintString;
PROCEDURE DoPrint;

IMPLEMENTATION

PROCEDURE SetFont( theFont : INTEGER; theSize : INTEGER; theStyle : Style );
	BEGIN
		TextFont( theFont );
		TextSize( theSize );
		TextFace( theStyle );
	END;


FUNCTION GetSelectedFonts : IntegerListRef;
	VAR
		head			: IntegerListRef;
		cur			: IntegerListRef;
		prev			: IntegerListRef;

		rowNum		: INTEGER;	
		myCell		: Cell;
		
		selected		: BOOLEAN;
		theData		: Str255;
		dataLen		: INTEGER;
	BEGIN
		
		head:= IntegerListRef( NewPtr( SizeOf( IntegerList ) ) );
		cur:= IntegerListRef( NewPtr( SizeOf( IntegerList ) ) );
		prev:= IntegerListRef( NewPtr( SizeOf( IntegerList ) ) );
		
		head^.next:= NIL;
		cur:=head;
		prev:=head;
		
		FOR rowNum:= 0  TO  gDoc^.docList^^.dataBounds.bottom DO BEGIN
			SetPt( myCell, 0 , rowNum );
			selected:= LGetSelect( FALSE, myCell, gDoc^.docList );
			
			IF ( selected ) THEN BEGIN
				dataLen:=255;
				LGetCell( @theData[1], dataLen , myCell, gDoc^.docList );
				theData[0]:= CHAR( dataLen );
				
				prev:=cur;
				cur:= IntegerListRef( NewPtr( SizeOf( IntegerList ) ) );
				cur^.Next:= NIL;
				prev^.Next:= cur;
				
				GetFNum( theData , cur^.num );
			END;
		END;
	
		GetSelectedFonts:=head^.next;
		
		head:=NIL;
		cur:=NIL;
		prev:=NIL;
	
		DisposePtr( Ptr( head ) );
		DisposePtr( Ptr( cur ) );
		DisposePtr( Ptr( prev ) );
	
	END;

FUNCTION NoMoreRoom( srcRect : Rect;  height : INTEGER ) : BOOLEAN;
	BEGIN
	
		IF height > ( srcRect.bottom - srcRect.top ) THEN
			NoMoreRoom:=TRUE
		ELSE NoMoreRoom:=FALSE;		
	
	END;

FUNCTION FontHeight : INTEGER;
	VAR
		fontData		: FontInfo;
	BEGIN
		GetFontInfo( fontData );
		FontHeight:= ( fontData.ascent + fontData.descent + fontData.leading );
	END;

FUNCTION NumberOfRows( thePage : Rect ) : INTEGER;
	VAR
		 length			: INTEGER;
		 index			: INTEGER;
		
		insert			: INTEGER;
		
		 width			: INTEGER;
		 rows			: INTEGER;
	BEGIN
		length:= ord( gPrintString[0] );
		insert:= thePage.left;
		
		rows:=1;
		
		FOR index:=1 TO length DO BEGIN
			width:= CharWidth( gPrintString[index] );
			IF ( insert + width + 5 ) > ( thePage.right - 10 ) THEN BEGIN
				insert:= thePage.left;
				rows:= ( rows + 1 );
			END
			
			ELSE BEGIN
				insert:= ( insert + width + 5 );
			END;
		END;
		
		NumberOfRows:= rows;		
		
	END;

PROCEDURE DrawFont( thePort : TPPrPort; VAR thePage : Rect; theFont : INTEGER );
	VAR
		insert			: INTEGER;
		height			: INTEGER;
		
		fontName			: Str255;
		fontID			: Str255;
		
		length			: INTEGER;
		width			: INTEGER;
		index			: INTEGER;
	BEGIN
		
		SetPort( @thePort^.gPort );
		insert:= thePage.left;
		
		{ 1. Draw the heading }
			SetFont( geneva, 9 , [bold] );
			height:= FontHeight;
			MoveTo( insert, ( height + thePage.top + 10 ) - 2 );
			GetFontName( theFont , fontName );
			NumToString( theFont, fontID );
			
			DrawString( ' Font: ' );
			TextFace( [] );
			DrawString( fontName );
			TextFace( [bold] );
			DrawString( ' ID: ');
			TextFace( [] );
			DrawString( fontID );
			
			MoveTo( insert, ( height + thePage.top + 10 ) );
			LineTo( thePage.right, ( height + thePage.top + 10 ) );
		
		{ 2. Draw in the characters of the font }
			length:= ORD( gPrintString[0] );
			insert:= thePage.left;
			
			SetFont( theFont, 12 , [] );
			height:= height + thePage.top + FontHeight + 10 ;
			
			FOR index:=1 TO length DO BEGIN
				width:= CharWidth( gPrintString[index] );
				IF ( insert + width + 5 ) > ( thePage.right -10 ) THEN BEGIN
					insert:= thePage.left;
					height:= height + FontHeight;
				END
			
				ELSE BEGIN
					insert:= ( insert + width + 5 );
					MoveTo( insert, height );
					DrawChar( gPrintString[index] );
				END;
			END;
			
		thePage.top:= ( thePage.top + ( height - thePage.top ) );
		
	END;

PROCEDURE ImagePage( thePort : TPPrPort; thePage : Rect; VAR fonts : IntegerListRef );
	VAR
		currentFont			: IntegerListRef;
		currentHeight			: INTEGER;
		headerHeight			: INTEGER;
		fontData				: FontInfo;
		
		PageFull				: BOOLEAN;
	BEGIN
	
		currentFont:= fonts;
		PageFull:= FALSE;		
		
		REPEAT
			
			SetFont( geneva, 12, [bold] );
			headerHeight:= FontHeight;
			
			SetFont( fonts^.num , 12 , [] );
			currentHeight:= FontHeight * ( NumberOfRows( thePage ) );			
			PageFull:= NoMoreRoom( thePage , ( currentHeight + headerHeight ) );
			
			IF NOT ( PageFull ) THEN BEGIN
				DrawFont( thePort , thePage, fonts^.num );
				fonts:= fonts^.next;
			END;			
			
		UNTIL PageFull OR ( fonts = NIL );
			
	END;

PROCEDURE PrintEngine( documentNum, pageNum : INTEGER; fonts : IntegerListRef );
	VAR
		thePort			: TPPrPort;
		thePage			: Rect;
		
		curPageNum		: INTEGER;
		status			: TPrStatus;		
	BEGIN
	
		IF ( documentNum < 0 ) THEN BEGIN
			
			REPEAT
				
				thePort:= PrOpenDoc( gPrintRef , NIL , NIL );
				
					thePage:= gPrintRef^^.prInfoPT.rPage;
					
					IF ( PrError = noErr ) THEN BEGIN
						PrOpenPage( thePort , NIL );
						IF ( PrError = noErr ) THEN BEGIN
							ImagePage( thePort , thePage , fonts );
						END;
						PrClosePage( thePort );
					END;

				PrCloseDoc( thePort );
				
				IF gPrintRef^^.prJob.bjDocLoop = bSpoolLoop THEN
					BEGIN
						IF PrError = noErr THEN PrPicFile( gPrintRef , NIL, NIL , NIL, status );
					END;

				
			UNTIL ( fonts = NIL );
			
		END;
		
		IF ( documentNum > 0 ) THEN BEGIN
			
			FOR curPageNum:=1 TO documentNum DO BEGIN
				
				thePort:= PrOpenDoc( gPrintRef , NIL , NIL );
				REPEAT					
					
					thePage:= gPrintRef^^.prInfoPT.rPage;
					
					IF ( PrError  = noErr ) THEN BEGIN
						PrOpenPage( thePort , NIL );
						IF ( PrError = noErr ) THEN BEGIN
							ImagePage( thePort , thePage , fonts );
						END;
						PrClosePage( thePort );
					END;
				UNTIL ( fonts  = NIL );
				PrCloseDoc( thePort );

				IF gPrintRef^^.prJob.bjDocLoop = bSpoolLoop THEN
					BEGIN
						IF PrError = noErr THEN PrPicFile( gPrintRef , NIL, NIL , NIL, status );
					END;
								
			END;
			
		END;
	
	END;

PROCEDURE GetPrintString;
	CONST
		kAllChars			= 1;
		kLettersNumbers	= 2;
		kLetters			= 3;
		kCustom			= 5;
		
	VAR
		len				: INTEGER;
		index			: INTEGER;
		currentChar		: INTEGER;
	BEGIN
	
		CASE gSettings^^.what OF
			kAllChars:  { 33 - 216 }
				BEGIN
					len:= ( 216 - 33 ) + 1;
					gPrintString[0]:= chr( len );
					
					currentChar:= 33;
					
					FOR index:= 1 TO len DO BEGIN
						gPrintString[index]:= chr( currentChar );
						currentChar:= ( currentChar + 1 );
					END;
				END;
				
			kLettersNumbers: { 48 - 57 , 65 - 90 , 97 - 122  }
				BEGIN
					len:= ( ( 57 - 48 ) + 1 ) + ( ( 90 - 65 ) + 1 ) + ( ( 122 - 97) + 1 );
					gPrintString[0]:= chr( len );
					
					currentChar:= 48;

					FOR index:= 1 TO  ( 57 - 48 ) + 1  DO BEGIN
						gPrintString[index]:= chr( currentChar );
						currentChar:= ( currentChar + 1 );
					END;
					
					currentChar:= 65;
					FOR index:=11 TO ( ( 90 - 65 ) + 1 ) + 10 DO BEGIN
						gPrintString[index]:= chr( currentChar );
						currentChar:= ( currentChar + 1 );
					END;
					
					currentChar:= 97;
					FOR index:= 37 TO ( ( 122 - 97 ) + 1 ) + 36 DO BEGIN
						gPrintString[index]:= chr( currentChar );
						currentChar:= ( currentChar + 1 );
					END;
					
				END;
				
			kLetters:	{ 65 - 90 , 97 - 122 }
				BEGIN
					len:= ( ( 90 - 65 ) + 1 ) + ( ( 122 - 97 ) + 1 );
					gPrintString[0]:= chr( len );
					
					currentChar:= 65;

					FOR index:= 1 TO ( 90 - 65 ) + 1 DO BEGIN
						gPrintString[index]:= chr( currentChar );
						currentChar:= ( currentChar + 1 );
					END;
					
					currentChar:=97;
					FOR index:= 27 TO ( ( 122 - 97 ) + 1 ) + 26 DO BEGIN
						gPrintString[index]:= chr( currentChar );
						currentChar:= ( currentChar + 1 );
					END;
				END;
				
			kCustom:
				BEGIN
					gPrintString:=gSettings^^.custom;
				END;
		END;
	
	END;

PROCEDURE DoPrint;
	VAR
		printerOK			: BOOLEAN;
		fonts				: IntegerListRef;
		
		copies			: INTEGER;
		
	BEGIN
	
		{ 1. Ensure that the current settings are alright - if not they are set to default }
		printerOK:= PrValidate( gPrintRef );	
		
		{ 2. Display a printing setup dialog }
		printerOK:= PrJobDialog( gPrintRef );
				
		IF ( printerOK ) THEN
				BEGIN
				
				{ 3. Load all of the fonts into an array of INTEGERs containing the font ID numbers }
					fonts:= IntegerListRef( NewPtr( SizeOf( IntegerList ) ) );
					fonts:=GetSelectedFonts;
				
				{ 3.5 Load the string into memory of what is going to be printed }
					GetPrintString;
					
				{ 4. Determine the printing method and act accordingly }
				
					CASE gSettings^^.how OF
						1:	{ Single Document }
							BEGIN
								PrintEngine( 1, -1 , fonts );
							END;
							
						2:	{ Individual Pages }
							BEGIN
								PrintEngine( -1, 1 , fonts );			
							END;
					END;
				
				END;
		
	END;

END.