unit SpriteWorld;

interface

	uses
{$IFC undefined THINK_Pascal}
		Types, Quickdraw, 
{$ENDC}
		SWCommonHeaders, SpriteLayer, SpriteFrame, Sprite, Tiling;

{$PUSH}
{$ALIGN MAC68K}

{/--------------------------------------------------------------------------------------}
{	SpriteWorld.h}
{}
{	Portions are copyright:  1991-94 Tony Myles, All rights reserved worldwide.}
{}
{	Description:	constants, structures, and prototypes for sprite worlds}
{/--------------------------------------------------------------------------------------}

{/--------------------------------------------------------------------------------------}
{	sprite world error constants}
{/--------------------------------------------------------------------------------------}

	const
		kSystemTooOldErr = 100;		{ < System 7.0}
		kMaxFramesErr = 101;				{ attempt to exceed maximum number of frames for a sprite}
		kInvalidFramesIndexErr = 102;		{ frame index out of range}
		kNotCWindowErr = 103;				{ attempt to make a SpriteWorld from non-color WindowPtr}
		kNilParameterErr = 104;			{ nil SpritePtr, FramePtr, etc.}
		kWrongDepthErr = 105;				{ invalid pixel size for attempted function}
		kWrongMaskErr = 106;				{ invalid mask type for attempted function}
		kOutOfRangeErr = 107;				{ tileID or tileMap value out of bounds}
		kTilingNotInitialized = 108;		{ tiling hasn't been initialized}
		kTilingAlreadyInitialized = 109;	{ tiling already initialized; can't be initialized again}
		kNullTileMapErr = 110;				{ no TileMap has ever been created/loaded}


{/--------------------------------------------------------------------------------------}
{	sprite world type definitions}
{/--------------------------------------------------------------------------------------}

	type
		SpriteWorldPeek = ^SpriteWorldRecX;
{$IFC undefined THINK_Pascal}
		WorldMoveProcPtr = procedure ( spriteWorldP : SpriteWorldPtr; followSpriteP: SpritePtr );
{$ELSEC}
		WorldMoveProcPtr = ProcPtr;
{$ENDC}
		XIntegerArray = array[0..0] of Integer;
		XIntegerArrayPtr = ^XIntegerArray;

{/--------------------------------------------------------------------------------------}
{	sprite world data structure}
{/--------------------------------------------------------------------------------------}

		SpriteWorldRecX = record
				headSpriteLayerP: SpriteLayerPtr;	{ head of the sprite layer linked list}
				tailSpriteLayerP: SpriteLayerPtr;	{ tail of the sprite layer linked list}

				backFrameP: FramePtr;				{ frame for the background}
				workFrameP: FramePtr;				{ work, or "mixing" frame}
				windowFrameP: FramePtr;				{ frame for drawing to the screen}

				offscreenDrawProc: DrawProcPtr;		{ callback for erasing sprites offscreen}
				screenDrawProc: DrawProcPtr;			{ callback for drawing sprite pieces onscreen}

				windRect: Rect;					{ holds windowFrameP->frameRect for easier access}
				backRect: Rect;					{ holds backFrameP->frameRect for easier access}

				visScrollRect: Rect;				{ rect that is copied to screen when scrolling}
				oldVisScrollRect: Rect;			{ visScrollRect from last frame}
				offscreenScrollRect: Rect;		{ same as visScrollRect, but local to offscreen}
				horizScrollRectOffset: Integer;		{ offset from offscreenScrollRect to visScrollRect}
				vertScrollRectOffset: Integer;		{ offset from offscreenScrollRect to visScrollRect}
				horizScrollDelta: Integer;			{ horizontal scrolling delta}
				vertScrollDelta: Integer;			{ vertical scrolling delta}
				scrollRectMoveBounds: Rect;		{ move bounds for visScrollRect}

				worldMoveProc: WorldMoveProcPtr;		{ pointer to the scrolling world move procedure}
				followSpriteP: SpritePtr;			{ pointer to the "follow sprite", or NULL}

				tileMap: TileMapPtr;				{ two-dimensional tile map array}
				tilingIsInitialized: Boolean;	{ has the tiling been initialized yet?}
				numTileMapRows: Integer;			{ number of rows in tileMap array}
				numTileMapCols: Integer;			{ number of cols in tileMap array}
				offscreenTileMap: IntegerPtr;		{ offscreen tile map array}
				numOffscreenTileMapRows: Integer;{ number of rows in offscreenTileMap array}
				numOffscreenTileMapCols: Integer;{ number of cols in offscreenTileMap array}
				tileFrameArray: FrameArrayPtr;		{ array of tile framePtrs}
				curTileImage: XIntegerArrayPtr;			{ array specifying the current frame of each tile}
				maxNumTiles: Integer;			{ number of elements in tileFrameArray}
				tileWidth: Integer;				{ width of each tile}
				tileHeight: Integer;				{ height of each tile}
				tilingIsOn: Boolean;				{ are the tiling routines turned on?}
				numTilesChanged: Integer;		{ number of rects in updateRect array to update}
				changedTiles: ^Rect;			{ array of rects of tiles that changed}
				tileChangeProc: TileChangeProcPtr;	{ pointer to tile frame changing procedure}
				tileMaskDrawProc: DrawProcPtr;		{ drawProc for drawing masked tiles above sprites}

				mainSWGDH: GDHandle;				{ GDH of SpriteWorld's window}
				pixelDepth: Integer;				{ SpriteWorld's depth }

				runningTimeCount: LongInt;	{ running total time in milliseconds}
				lastMicroseconds: UnsignedWide;	{ value of previous Microseconds() call}
				frameHasOccured: Boolean;	{ Has the SpriteWorld been processed?}
				fpsTimeInterval: Integer;	{ milliseconds per frame of animation (1000/fps)}
				timeOfLastFrame: LongInt;	{ time (from runningTimeCount) of last frame}
				vblTaskRec: VBLTaskRec;			{ extended VBLTask record}
				usingVBL: Boolean;			{ is the VBL task installed?}

				anySpritesNeedRemoving: Boolean;		{ flag that at least one Sprite needs to be removed}

				userData: LongInt;						{ reserved for user}
			end;


	function SWEnterSpriteWorld: OSErr;
	procedure SWExitSpriteWorld;

	function SWCreateSpriteWorld (mainGDH: GDHandle; var spriteWorldP: SpriteWorldPtr; windowFrameP: FramePtr; backFrameP: FramePtr; workFrameP: FramePtr): OSErr;
	function SWCreateSpriteWorldFromWindow (var spriteWorldP: SpriteWorldPtr; srcWindowP: CWindowPtr; worldRectP: RectPtr; offscreenRectP: RectPtr): OSErr;
	procedure SWDisposeSpriteWorld (spriteWorldP: SpriteWorldPtr);

	procedure SWAddSpriteLayer (spriteWorldP: SpriteWorldPtr; spriteLayerP: SpriteLayerPtr);
	procedure SWRemoveSpriteLayer (spriteWorldP: SpriteWorldPtr; spriteLayerP: SpriteLayerPtr);
	procedure SWSwapSpriteLayer (spriteWorldP: SpriteWorldPtr; srcSpriteLayerP: SpriteLayerPtr; dstSpriteLayerP: SpriteLayerPtr);
	function SWGetNextSpriteLayer (spriteWorldP: SpriteWorldPtr; curSpriteLayerP: SpriteLayerPtr): SpriteLayerPtr;

	procedure SWLockSpriteWorld (spriteWorldP: SpriteWorldPtr);
	procedure SWUnlockSpriteWorld (spriteWorldP: SpriteWorldPtr);
	procedure SWSetPortToBackground (spriteWorldP: SpriteWorldPtr);
	procedure SWSetPortToWorkArea (spriteWorldP: SpriteWorldPtr);
	procedure SWSetPortToWindow (spriteWorldP: SpriteWorldPtr);
	function SWSetSpriteWorldOffscreenDrawProc (spriteWorldP: SpriteWorldPtr; offscreenProc: DrawProcPtr): OSErr;
	function SWSetSpriteWorldScreenDrawProc (spriteWorldP: SpriteWorldPtr; drawProc: DrawProcPtr): OSErr;
	procedure SWStdWorldDrawProc (srcFrameP: FramePtr; dstFrameP: FramePtr; var srcRect: Rect; var dstRect: Rect);

	procedure SWCopyBackgroundToWorkArea (spriteWorldP: SpriteWorldPtr);
	procedure SWUpdateWindow (spriteWorldP: SpriteWorldPtr);
	procedure SWUpdateSpriteWorld (spriteWorldP: SpriteWorldPtr; updateWindow: Boolean);
	procedure SWProcessSpriteWorld (spriteWorldP: SpriteWorldPtr);
	procedure SWAnimateSpriteWorld (spriteWorldP: SpriteWorldPtr);

	procedure SWCheckIdleSpriteOverlap (spriteWorldP: SpriteWorldPtr; idleSpriteP: SpritePtr; headActiveSpriteP: SpritePtr);
	procedure SWFindSpritesToBeRemoved (spriteWorldP: SpriteWorldPtr);
	procedure SWSetSpriteWorldMaxFPS (spriteWorldP: SpriteWorldPtr; framesPerSec: Integer);
	function SWSyncSpriteWorldToVBL (spriteWorldP: SpriteWorldPtr; syncingOn: Boolean): OSErr;

	function SWGetSpriteWorldVersion: LongInt;

{$ifc GENERATINGPOWERPC}
	procedure SWVBLTask (vblTaskPtr: VBLTaskRecPtr);(* Under CFM, we're passed the VBL task pointer *)
{$elsec}
	procedure SWVBLTask;					(* Otherwise, we'll have to get it out of register A0 *)
{$endc}

{$ALIGN RESET}
{$POP}

{$IFC not undefined THINK_Pascal}
implementation
{$ENDC}
end.
